from urllib.parse import SplitResult
import mxdevtool as mx
import mxdevtool.utils as utils
import mxdevtool.marketconvension as mx_m
from typing import List


def setSmithWilsonParameter(curve: mx.YieldCurveExt, d):
    extrapolationType = mx.extrapolator1DParse(d['extrapolationType'])

    if extrapolationType == mx.Extrapolator1D.SmithWilson:
        alpha = d['smithwilsonAlpha']
        ufr = d['smithwilsonUFR']
        curve.setSmithwilsonParameter(alpha, ufr)

    return curve


# termstructures
class FlatForward(mx.core_FlatForward):
    def __init__(self, refDate: mx.Date, forward: float,
                   calendar=mx.NullCalendar(),
                   dayCounter=mx.Actual365Fixed(),
                   businessDayConvention=mx.ModifiedFollowing,
                   compounding=mx.Compounded):

        self._refDate = utils.toDateCls(refDate)
        self._forward = forward
        self._calendar = utils.toCalendarCls(calendar)
        self._dayCounter = utils.toDayCounterCls(dayCounter)
        self._businessDayConvention = utils.toBusinessDayConvention(businessDayConvention)
        self._compounding = utils.toCompounding(compounding)

        mx.core_FlatForward.__init__(self, self._refDate, forward, calendar, dayCounter, businessDayConvention, compounding)

    @staticmethod
    def fromDict(d: dict, mrk=mx.MarketData()):
        mx.check_fromDict(d, mx.CLASS_TYPE_NAME, FlatForward.__name__)

        refDate = utils.toDateCls(d['refDate'])
        forward = d['forward']
        calendar = utils.toCalendarCls(d['calendar'])
        dayCounter = utils.toDayCounterCls(d['dayCounter'])
        businessDayConvention = utils.toBusinessDayConvention(d['businessDayConvention'])
        compounding = utils.toCompounding(d['compounding'])

        return FlatForward(refDate, forward, calendar, dayCounter, businessDayConvention, compounding)

    def toDict(self):
        res = dict()

        res[mx.CLASS_TYPE_NAME] = self.__class__.__name__
        res['refDate'] = str(self._refDate)
        res['forward'] = self._forward
        res['calendar'] = str(self._calendar)
        res['dayCounter'] = str(self._dayCounter)
        res['businessDayConvention'] = mx.businessDayConventionToString(self._businessDayConvention)
        res['compounding'] = str(self._compounding)

        return res

    def shocked_clone(self, shock, method='absolute'):
        v = self._forward
        if method == 'relative': v *= (1.0 + shock)
        else: v += shock

        return FlatForward(self._refDate, v, self._calendar, self._dayCounter, self._businessDayConvention, self._compounding)

    def clone(self, **kwargs):
        refDate = kwargs['refDate'] if 'refDate' in kwargs else self._refDate
        forward = kwargs['forward'] if 'forward' in kwargs else self._forward
        calendar = kwargs['calendar'] if 'calendar' in kwargs else self._calendar
        dayCounter = kwargs['dayCounter'] if 'dayCounter' in kwargs else self._dayCounter
        businessDayConvention = kwargs['businessDayConvention'] if 'businessDayConvention' in kwargs else self._businessDayConvention
        compounding = kwargs['compounding'] if 'compounding' in kwargs else self._compounding

        return FlatForward(refDate, forward, calendar, dayCounter, businessDayConvention, compounding)

    def graph_view(self, typ=None, **kwargs):
        from mxdevtool.view import graph

        # period ( 1Y to 20Y )
        x =  [ v+1 for v in range(20) ]
        y1 = [ self.zeroRate(t, self._compounding).rate() for t in x]

        ydata_d = {'zeroRates': y1}
        return graph(name='FlatForward', xdata=x, ydata_d=ydata_d, **kwargs)


class ZeroYieldCurve(mx.core_ZeroYieldCurveExt):
    def __init__(self, refDate: mx.Date, periods: List[str], zeroRates: List[float],
                   interpolationType=mx.Interpolator1D.Linear,
                   extrapolationType=mx.Extrapolator1D.FlatForward,
                   calendar=mx.NullCalendar(),
                   dayCounter=mx.Actual365Fixed(),
                   businessDayConvention=mx.ModifiedFollowing,
                   compounding=mx.Compounded):

        self._refDate = utils.toDateCls(refDate)
        self._periods = utils.toStrList(periods)
        self._zeroRates = zeroRates
        self._interpolationType = utils.toInterpolation1DType(interpolationType)
        self._extrapolationType = utils.toExtrapolation1DType(extrapolationType)
        self._calendar = utils.toCalendarCls(calendar)
        self._dayCounter = utils.toDayCounterCls(dayCounter)
        self._businessDayConvention = utils.toBusinessDayConvention(businessDayConvention)
        self._compounding = utils.toCompounding(compounding)

        mx.core_ZeroYieldCurveExt.__init__(self, self._refDate, self._periods, zeroRates, interpolationType, extrapolationType,
                                    self._calendar, self._dayCounter, self._businessDayConvention, self._compounding)

    @staticmethod
    def fromDict(d: dict, mrk=mx.MarketData()):
        mx.check_fromDict(d, mx.CLASS_TYPE_NAME, ZeroYieldCurve.__name__)

        refDate = d['refDate']
        periods = d['periods']
        zeroRates = d['zeroRates']
        interpolationType = utils.toInterpolation1DType(d['interpolationType'])
        extrapolationType = utils.toExtrapolation1DType(d['extrapolationType'])
        calendar = d['calendar']
        dayCounter = d['dayCounter']
        businessDayConvention = d['businessDayConvention']
        compounding = d['compounding']

        curve = ZeroYieldCurve(refDate, periods, zeroRates, interpolationType, extrapolationType,
                               calendar, dayCounter, businessDayConvention, compounding)

        setSmithWilsonParameter(curve, d)

        return curve

    def toDict(self):
        res = dict()

        res[mx.CLASS_TYPE_NAME] = self.__class__.__name__
        res['refDate'] = str(self._refDate)
        res['periods'] = self._periods
        res['zeroRates'] = self._zeroRates
        res['interpolationType'] = mx.interpolator1DToString(self._interpolationType)
        res['extrapolationType'] = mx.extrapolator1DToString(self._extrapolationType)
        res['calendar'] = str(self._calendar)
        res['dayCounter'] = str(self._dayCounter)
        res['businessDayConvention'] = mx.businessDayConventionToString(self._businessDayConvention)
        res['compounding'] = mx.compoundingToString(self._compounding)

        return res

    def shocked_clone(self, shock, method='absolute'):
        shocked_value = utils.calculateShock(shock, method, self._vol)

        return ZeroYieldCurve(self._refDate, shocked_value, self._calendar, self._dayCounter, self._businessDayConvention, self._compounding)

    def clone(self, **kwargs):
        refDate = kwargs['refDate'] if 'refDate' in kwargs else self._refDate
        periods = kwargs['periods'] if 'periods' in kwargs else self._periods
        zeroRates = kwargs['zeroRates'] if 'zeroRates' in kwargs else self._zeroRates
        interpolationType = kwargs['interpolationType'] if 'interpolationType' in kwargs else self._interpolationType
        extrapolationType = kwargs['extrapolationType'] if 'extrapolationType' in kwargs else self._extrapolationType
        calendar = kwargs['calendar'] if 'calendar' in kwargs else self._calendar
        dayCounter = kwargs['dayCounter'] if 'dayCounter' in kwargs else self._dayCounter
        businessDayConvention = kwargs['businessDayConvention'] if 'businessDayConvention' in kwargs else self._businessDayConvention
        compounding = kwargs['compounding'] if 'compounding' in kwargs else self._compounding

        return ZeroYieldCurve(refDate, periods, zeroRates, interpolationType, extrapolationType, calendar, dayCounter, businessDayConvention, compounding)


    def graph_view(self, **kwargs):
        from mxdevtool.view import graph

        x = [ p.yearFraction() for p in utils.toPeriodClsList(self._periods)]
        y1 = self._zeroRates

        ydata_d = {'zeroRates': y1}
        return graph(name='ZeroYieldCurve', xdata=x, ydata_d=ydata_d, **kwargs)


# familyname : irskrw_krccp only now
class BootstapSwapCurveCCP(mx.core_BootstapSwapCurveCCP):
    def __init__(self, refDate, periods, rateTypes, quotes,
                interpolationType=mx.Interpolator1D.Linear,
                extrapolationType=mx.Extrapolator1D.FlatForward,
                familyname='irskrw_krccp',
                forSettlement=True):

        self._refDate = utils.toDateCls(refDate)
        self._periods = utils.toStrList(periods)
        self._rateTypes = rateTypes
        self._quotes = quotes
        self._interpolationType = utils.toInterpolation1DType(interpolationType)
        self._extrapolationType = utils.toExtrapolation1DType(extrapolationType)
        self._familyname = familyname
        self._forSettlement = forSettlement

        extrapolation = utils.toExtrapolationCls(extrapolationType)

        mx.core_BootstapSwapCurveCCP.__init__(self, self._refDate, self._periods, rateTypes, quotes,
                                              interpolationType, extrapolation,
                                              familyname, forSettlement)

    @staticmethod
    def fromDict(d: dict, mrk=mx.MarketData()):
        mx.check_fromDict(d, mx.CLASS_TYPE_NAME, BootstapSwapCurveCCP.__name__)

        refDate = d['refDate']
        periods = d['periods']
        rateTypes = d['rateTypes']
        quotes = d['quotes']
        interpolationType = utils.toInterpolation1DType(d['interpolationType'])
        extrapolationType = utils.toExtrapolation1DType(d['extrapolationType'])
        familyname = d['familyname']
        forSettlement = d['forSettlement']

        curve = BootstapSwapCurveCCP(refDate, periods, rateTypes, quotes,
                                     interpolationType, extrapolationType,
                                     familyname, forSettlement)

        setSmithWilsonParameter(curve, d)

        return curve

    def toDict(self):
        res = dict()

        res[mx.CLASS_TYPE_NAME] = self.__class__.__name__

        res['refDate'] = str(self._refDate)
        res['periods'] = self._periods
        res['rateTypes'] = self._rateTypes
        res['quotes'] = self._quotes
        res['interpolationType'] = mx.interpolator1DToString(self._interpolationType)
        res['extrapolationType'] = mx.extrapolator1DToString(self._extrapolationType)

        if self._extrapolationType == mx.Extrapolator1D.SmithWilson:
            res['smithwilsonAlpha'] = self.smithwilsonAlpha()
            res['smithwilsonUFR'] = self.smithwilsonUFR()

        res['familyname'] = self._familyname
        res['forSettlement'] = self._forSettlement

        return res

    def shocked_clone(self, shock, method='absolute'):
        shocked_value = utils.calculateShock(shock, method, self._quotes)

        return BootstapSwapCurveCCP(self._refDate, self._periods, self._rateTypes, shocked_value, 
                                    self._interpolationType, self._extrapolationType, self._familyname, self._forSettlement)

    def clone(self, **kwargs):
        refDate = kwargs['refDate'] if 'refDate' in kwargs else self._refDate
        periods = kwargs['periods'] if 'periods' in kwargs else self._periods
        rateTypes = kwargs['rateTypes'] if 'rateTypes' in kwargs else self._rateTypes
        quotes = kwargs['quotes'] if 'quotes' in kwargs else self._quotes
        interpolationType = kwargs['interpolationType'] if 'interpolationType' in kwargs else self._interpolationType
        extrapolationType = kwargs['extrapolationType'] if 'extrapolationType' in kwargs else self._extrapolationType

        familyname = kwargs['familyname'] if 'familyname' in kwargs else self._familyname
        forSettlement = kwargs['forSettlement'] if 'forSettlement' in kwargs else self._forSettlement

        curve = BootstapSwapCurveCCP(refDate, periods, rateTypes, quotes, interpolationType, extrapolationType, familyname, forSettlement)

        smithwilsonAlpha = kwargs['smithwilsonAlpha'] if 'smithwilsonAlpha' in kwargs else self.smithwilsonAlpha()
        smithwilsonUFR = kwargs['smithwilsonUFR'] if 'smithwilsonUFR' in kwargs else self.smithwilsonUFR()
        curve.setSmithwilsonParameter(smithwilsonAlpha, smithwilsonUFR)

        return curve


    def graph_view(self, **kwargs):
        from mxdevtool.view import graph

        x = [ p.yearFraction() for p in utils.toPeriodClsList(self._periods)]
        y1 = self._quotes

        ydata_d = {'quotes': y1}

        return graph(name='BootstapSwapCurveCCP', xdata=x, ydata_d=ydata_d, **kwargs)


# generic(cash, bond, swap)
class BootstrapCurve(mx.core_BootstrapCurve):
    def __init__(self, refDate: mx.Date, periods: List[str or mx.Period], rateTypes: List[str], quotes: List[float],
                interpolationType=mx.Interpolator1D.Linear,
                extrapolationType=mx.Extrapolator1D.FlatForward,
                marketConvension=mx_m.get_marketConvension_vanillaswap('irskrw')):
        '''
        Args:
            refDate (Date) : referenceDate
            rateTypes (List[str]): Cash, Swap, Bond
        '''
        self._refDate = utils.toDateCls(refDate)
        self._periods = utils.toStrList(periods)
        self._rateTypes = rateTypes # 
        self._quotes = quotes
        self._interpolationType = utils.toInterpolation1DType(interpolationType)
        self._extrapolationType = utils.toExtrapolation1DType(extrapolationType)
        self._marketConvension = marketConvension

        extrapolation = utils.toExtrapolationCls(extrapolationType)

        mx.core_BootstrapCurve.__init__(self, self._refDate, self._periods, rateTypes, quotes,
                                           interpolationType, extrapolation, marketConvension)

    @staticmethod
    def fromDict(d: dict, mrk=mx.MarketData()):
        mx.check_fromDict(d, mx.CLASS_TYPE_NAME, BootstrapCurve.__name__)

        refDate = d['refDate']
        periods = d['periods']
        rateTypes = d['rateTypes']
        quotes = d['quotes']
        interpolationType = utils.toInterpolation1DType(d['interpolationType'])
        extrapolationType = utils.toExtrapolation1DType(d['extrapolationType'])
        marketConvension = mx_m.marketConvensionFromDict(d['marketConvension'])

        curve = BootstrapCurve(refDate, periods, rateTypes, quotes,
                                    interpolationType, extrapolationType,
                                    marketConvension)

        setSmithWilsonParameter(curve, d)

        return curve


    def toDict(self):
        res = dict()

        res[mx.CLASS_TYPE_NAME] = self.__class__.__name__

        res['refDate'] = str(self._refDate)
        res['periods'] = self._periods
        res['rateTypes'] = self._rateTypes
        res['quotes'] = self._quotes
        res['interpolationType'] = mx.interpolator1DToString(self._interpolationType)
        res['extrapolationType'] = mx.extrapolator1DToString(self._extrapolationType)

        if self._extrapolationType == mx.Extrapolator1D.SmithWilson:
            res['smithwilsonAlpha'] = self.smithwilsonAlpha()
            res['smithwilsonUFR'] = self.smithwilsonUFR()

        res['marketConvension'] = self._marketConvension.toDict()
        
        return res

    def shocked_clone(self, shock, method='absolute'):
        shocked_value = utils.calculateShock(shock, method, self._quotes)

        return BootstrapCurve(self._refDate, self._periods, self._rateTypes, shocked_value, 
                                    self._interpolationType, self._extrapolationType, self._marketConvension)

    def clone(self, **kwargs):
        refDate = kwargs['refDate'] if 'refDate' in kwargs else self._refDate
        periods = kwargs['periods'] if 'periods' in kwargs else self._periods
        rateTypes = kwargs['rateTypes'] if 'rateTypes' in kwargs else self._rateTypes
        quotes = kwargs['quotes'] if 'quotes' in kwargs else self._quotes
        interpolationType = kwargs['interpolationType'] if 'interpolationType' in kwargs else self._interpolationType
        extrapolationType = kwargs['extrapolationType'] if 'extrapolationType' in kwargs else self._extrapolationType
        marketConvension = kwargs['marketConvension'] if 'marketConvension' in kwargs else self._marketConvension

        curve = BootstrapCurve(refDate, periods, rateTypes, quotes, interpolationType, extrapolationType, marketConvension)

        smithwilsonAlpha = kwargs['smithwilsonAlpha'] if 'smithwilsonAlpha' in kwargs else self.smithwilsonAlpha()
        smithwilsonUFR = kwargs['smithwilsonUFR'] if 'smithwilsonUFR' in kwargs else self.smithwilsonUFR()
        curve.setSmithwilsonParameter(smithwilsonAlpha, smithwilsonUFR)

        return curve

    def graph_view(self, **kwargs):
        from mxdevtool.view import graph

        x = [ p.yearFraction() for p in utils.toPeriodClsList(self._periods)]
        y1 = self._quotes

        ydata_d = {'quotes': y1}

        return graph(name='BootstrapCurve', xdata=x, ydata_d=ydata_d, **kwargs)



class ZeroSpreadedCurve(mx.core_ZeroSpreadedTermStructure):
    def __init__(self, baseCurve, spread, compounding=mx.Continuous, frequency=mx.Annual):

        self._baseCurve = baseCurve
        self._spread = spread
        self._compounding = compounding
        self._frequency = frequency

        mx.core_ZeroSpreadedTermStructure.__init__(self, baseCurve, spread, compounding, frequency)

    @staticmethod
    def fromDict(d: dict, mrk=mx.MarketData()):
        mx.check_fromDict(d, mx.CLASS_TYPE_NAME, ZeroSpreadedCurve.__name__)

        baseCurve = mrk.get_yieldCurve_d(d['baseCurve'])
        spread = d['spread']
        compounding = mx.compoundingParse(d['compounding'])
        frequency = mx.frequencyParse(d['frequency'])

        return ZeroSpreadedCurve(baseCurve, spread, compounding, frequency)

    def toDict(self):
        res = dict()

        res[mx.CLASS_TYPE_NAME] = self.__class__.__name__

        res['baseCurve'] = self._baseCurve.toDict()
        res['spread'] = self._spread
        res['compounding'] = self._compounding
        res['frequency'] = self._frequency

        return res

    # shock to spread
    def shocked_clone(self, shock, method='absolute'):
        shocked_value = utils.calculateShock(shock, method, self._spread)

        return ZeroSpreadedCurve(self._baseCurve, shocked_value, self._compounding, self._frequency)

    def clone(self, **kwargs):
        baseCurve = kwargs['baseCurve'] if 'baseCurve' in kwargs else self._baseCurve
        spread = kwargs['spread'] if 'spread' in kwargs else self._spread
        compounding = kwargs['compounding'] if 'compounding' in kwargs else self._compounding
        frequency = kwargs['frequency'] if 'frequency' in kwargs else self._frequency

        return ZeroSpreadedCurve(baseCurve, spread, compounding, frequency)

    def graph_view(self, **kwargs):
        from mxdevtool.view import graph

        x = kwargs.get('x')

        # period ( 1Y to 20Y )
        x = [v+1 for v in range(20)] if x is None else x
        y1 = [self.zeroRate(t, self._compounding).rate() for t in x]
        ydata_d = {'zeroRates': y1}

        return graph(name='ZeroSpreadedCurve', xdata=x, ydata_d=ydata_d, **kwargs)


# class ForwardSpreadedCurve(mx.core_ForwardSpreadedTermStructure):
#     def __init__(self, curve, spread):
#         mx.core_ForwardSpreadedTermStructure.__init__(self, curve, spread)

